//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
//
//  Copyright(c)  2005  Richard B. Johnson
//
//  This program may be distributed under the GNU Public License
//  version 2, as published by the Free Software Foundation, Inc.,
//  59 Temple Place, Suite 330 Boston, MA, 02111.
//
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
/*
 *  Everything in this file is "secret".  I don't use the awful
 *  streams-based FILE object. I use my own simple structure.
 *  So, to keep this an opaque object, and to keep other code
 *  from doing the wrong thing, I do not include <stdio.h>.
 *  Therefore, anything that might be declared there must be
 *  declared here. The user, of course, includes <stdio.h> as
 *  usual.
 *
 *  As far as user-code is considered, the FILE object is always
 *  accessed as a pointer. The user doesn't care that the pointer
 *  points to something different than what is shown in the header
 *  files. Users are not supposed to access members of this structure
 *  directly, and if they do, it's their problem.
 *
 *  The most common 'f' text-based procedures are in this file.
 *  If you need more, just add them. If you need a fcloseall(),
 *  you will need to change the FILE object to have a forward and
 *  backward pointer and handle them, in every fopen() and fclose().
 */


#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <stdarg.h>
#include <wait.h>
#include <errno.h>
#include <string.h>
#include <malloc.h>
#include <fcntl.h>
extern void __set_errno(int);


#define CHECK(p)             \
    if(p == NULL)            \
    {                        \
        __set_errno(EINVAL); \
        return FAIL;         \
    }
#define NCHECK(p)            \
    if(p == NULL)            \
    {                        \
        __set_errno(EINVAL); \
        return NULL;         \
    }
#define VCHECK(p)            \
    if(p == NULL)            \
    {                        \
        __set_errno(EINVAL); \
        return;              \
    }
extern void reaper(int signal);
extern int __get_errno(void);
extern int printf(const char *,...);
extern int vsnprintf(char *buf, size_t len, const char *fmt, va_list va);
extern int vsscanf(char *buf, const char *fmt, va_list va);
#define BUF_LEN  0x1000
#define NORMAL   0xdead
#define POPEN    0xface
#define NR_ARGS  0x10
#define FAIL -1
#define EOF  -1

typedef struct {
    char buf[BUF_LEN];   /* Data buffer           */
    size_t pos;          /* Current position      */
    size_t io_len;       /* Remaining read length */
    int fd;              /* File descriptor       */
    int eof;             /* EOF if EOF encounter  */
    int err;             /* Saved errno on error  */
    pid_t pid;           /* For popen() child     */
    int pfd[2];          /* For popen pipes       */
    int magic;           /* NORMAL or POPEN       */
    } FILE;

extern int fprintf(FILE *, const char *,...);
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  These things are actually in start.S -- cute eh?
 */
extern FILE *stdin;
extern FILE *stdout;
extern FILE *stderr;
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  This is called from start.S, just before main() gets called.
 *  It sets up valid 'FILE' descriptors for stdin, stdout, and
 *  stderr. We allow these to be closed by setting 'magic' to
 *  NORMAL.
 */

void __setup_stdio()
{
    if((stdin = (FILE *) malloc(sizeof(FILE))) == NULL)
        return;
    bzero(stdin,  sizeof(FILE));
    stdin->fd    = STDIN_FILENO;
    stdin->magic = NORMAL;
    if((stdout = (FILE *) malloc(sizeof(FILE))) == NULL)
        return;
    bzero(stdout, sizeof(FILE));
    stdout->fd    = STDOUT_FILENO;
    stdout->magic = NORMAL;
    if((stderr = (FILE *) malloc(sizeof(FILE))) == NULL)
        return;
    bzero(stderr,  sizeof(FILE));
    stderr->fd    = STDERR_FILENO;
    stderr->magic = NORMAL;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  This appends a '\n' to a string and writes it to standard
 *  output.
 */
int puts(const char *cp)
{
    return printf("%s\n", cp);
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This writes a formatted string to standard output.
 */
int printf(const char *fmt,...)
{
    int ret;
    va_list va;
    va_start(va, fmt);
    ret = vsnprintf(stdout->buf, BUF_LEN, fmt, va);
    va_end(va);
    ret = write(STDOUT_FILENO, stdout->buf, ret);
    return ret;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  This is to open a 'stream' for buffered I/O. Check <stdio.h> for
 *  details.
 */
FILE *fopen(char *fname, char *how)
{
    int fd;
    int whence;
    FILE *file;

    fd = 0;
    whence = SEEK_SET;
    switch((int)*how)
    {
    case 'r':
        fd = O_RDONLY;
        if(how[1] == '+')
            fd = O_RDWR;
        break;
    case 'w':
        fd = O_RDWR|O_TRUNC|O_CREAT;
        if(how[1] == '+')
            fd |= O_CREAT;
        break;
    case 'a':
        whence = SEEK_END;
        fd = O_WRONLY|O_CREAT;
        if(how[1] == '+')
            fd = O_RDWR|O_CREAT;
        break;
    default:
        return NULL;
    }
    if((fd = open(fname, fd, 0644)) != FAIL)
    {
        file = (FILE *) malloc(sizeof(FILE));
        bzero(file,  sizeof(FILE));
        file->fd = fd;
        file->pos = lseek(fd, 0, whence);
        file->magic = NORMAL;
        return file;
    }
    return NULL;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  This associates an existing file descriptor with a FILE pointer
 *  for buffered I/O. Check <stdio.h> for
 *  details.
 */
FILE *fdopen(int fd, char *how)
{
    int mode;
    int flags;
    int whence;
    FILE *file;
    whence = SEEK_SET;        // Default, keep existing position
    switch((int)*how)
    {
    case 'r':
        mode = O_RDONLY;	// R/O
        if(how[1] == '+')       // Unless it has a +
             mode = O_RDWR;     // Then it is R/W
        break;
    case 'w':
        mode = O_RDWR;          // R/W
        break;
    case 'a':
        whence = SEEK_END;      // Append goes to end
        mode = O_RDWR;          // Mode is R/W
        break;
    default:
        return NULL;
    }
    if((flags = fcntl(fd, F_GETFL)) == FAIL)
        return NULL;

    if((flags & O_ACCMODE) != mode)
        fprintf(stderr, "Something wrong!\n");
    file = (FILE *) malloc(sizeof(FILE));
    bzero(file,  sizeof(FILE));
    file->fd = fd;
    file->pos = lseek(fd, 0, whence);
    file->magic = NORMAL;
    return file;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  This closes a 'stream' previously opened. Check <stdio.h> for details.
 */
int fclose(FILE *file)
{
    int ret;
    CHECK(file);
    if(file->magic != NORMAL)
    {
        __set_errno(EBADF);
        return FAIL;
    }
    file->eof = 0;
    ret = close(file->fd);
    free(file);
    return ret;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This performs a buffered, formatted write to a stream.
 *   check <stdio.h> for details.
 */
int fprintf(FILE *file, const char *fmt,...)
{
    va_list va;
    int ret;
    CHECK(file);
    va_start(va, fmt);
    ret = vsnprintf(file->buf, BUF_LEN, fmt, va);
    va_end(va);
    if((ret = write(file->fd, file->buf, ret)) == FAIL)
        file->err = __get_errno();
    return ret;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This writes formatted text to a file when a va_list already
 *   exists.
 */
int vfprintf(FILE *file, const char *fmt, va_list va)
{
    int ret;
    CHECK(file);
    ret = vsnprintf(file->buf, BUF_LEN, fmt, va);
    if((ret = write(file->fd, file->buf, ret)) == FAIL)
        file->err = __get_errno();
    return ret;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This reads a line at a time from the 'stream' specified in
 *   the argument. Check <stdio.h> for details.
 */
char *fgets(char *buf, size_t len, FILE *file)
{
    int ret;
    char *cp, c;
    NCHECK(file);
    file->eof = 0;
    cp = buf;
    while(len)
    {
        if(file->io_len == 0)  /* No bytes remain */
        {
            __set_errno(0x00);  // IEEE Std 1003.1-2001
            if((ret = read(file->fd, file->buf, BUF_LEN)) <= 0)
            {
                if((ret == -1) && (errno == EINTR))
                    continue;
                file->eof = EOF;
                if(cp == buf)
                    return NULL;
                *cp = (char) 0x00;
                return buf;
            }
            file->pos = 0;
            file->io_len = (size_t) ret;
        }
        c = file->buf[file->pos++];
        *cp++ = c;
        file->io_len--;
        len--;
        if(c == (char) '\n')
        {
            *cp = (char) 0x00;
            break;
        }
    }
    return buf;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This writes a string to the output stream. Check <stdio.h>
 *   for details.
 */
int fputs(const char *cp, FILE *file)
{
    size_t len;
    int ret;
    CHECK(file);
    len = strlen(strcpy(file->buf, cp));
    if((ret = write(file->fd, file->buf, len)) == FAIL)
        file->err = __get_errno();
    return ret;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  This writes a character to an output stream.
 *  Check <stdio.h> for details.
 */
int fputc(int c, FILE *file)
{
    CHECK(file);
    if(write(file->fd, (char *)&c, 0x01) == FAIL)
    {
        file->err = __get_errno();
        return EOF;
    }
    return c;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  This writes a character to the output stream.
 *  Check <stdio.h> for details.
 */
int putchar(int c)
{
    CHECK(stdout);
    if(write(stdout->fd, (char *)&c, 0x01) == FAIL)
    {
        stdout->err = __get_errno();
        return EOF;
    }
    return c;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This sets a stream to its start. Check <stdio.h> for details.
 */
void rewind(FILE *file)
{
    VCHECK(file);
    file->eof = 0;
    (void)lseek(file->fd, 0, SEEK_SET);
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This sets a stream to a position. Check <stdio.h> for defails.
 */
int fseek(FILE *file, long offset, int whence)
{
    CHECK(file);
    file->eof = 0;
    return lseek(file->fd, offset, whence);
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This gets the current position of a stream. Check <stdio.h>
 *   for details.
 */
long ftell(FILE *file)
{
    CHECK(file);
    return lseek(file->fd, 0, SEEK_CUR);
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This gets the current position of a stream.
 */
int fgetpos(FILE *file, long *pos)
{
    long ret;
    CHECK(file);
    if((ret = lseek(file->fd, 0, SEEK_CUR)) < 0)
        return ret;
    *pos = ret;
    return 0;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  This sets the current position for a stream.
 */
int fsetpos(FILE *file, long *pos)
{
    CHECK(file);
    file->eof = 0;
    return lseek(file->fd, *pos, SEEK_SET);
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This extracts parameters from a stream and writes them to
 *   variables based upon a format string.
 */
int fscanf(FILE *file, const char *fmt,...)
{
    char *buf;
    int i;
    va_list va;
    CHECK(file);
    file->eof = 0;
    if((buf = (char *) malloc(BUF_LEN * sizeof(char))) == NULL)
    {
        __set_errno(ENOMEM);
        return FAIL;
    }
    if(fgets(buf, BUF_LEN, file) == NULL)
    {
        free(buf);
        return FAIL;
    }
    va_start(va, fmt);
    i = vsscanf(buf, fmt, va);
    va_end(va);
    free(buf);
    return i;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This reads a number of objects of specified width from a
 *   stream. It returns the number of objects actually read.
 */
int fread(void *where, size_t width, size_t nr, FILE *file)
{
    int nr_read;
    CHECK(file);
    file->eof = 0;
    retry:;
    __set_errno(0x00);  // IEEE std 1003.1-2001
    if((nr_read = read(file->fd, where, nr * width)) <= 0)
    {
        if((nr_read == -1) && (errno == EINTR))
            goto retry;
        file->eof = EOF;
        return nr_read;
    }
    return nr_read / width;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This writes a number of objects of specified width to a
 *   stream. It returns the number of objects actually written.
 */
int fwrite(void *where, size_t width, size_t nr, FILE *file)
{
    int nr_write;
    CHECK(file);
    if((nr_write = write(file->fd, where, nr * width)) == FAIL)
    {
        file->err = __get_errno();
        return nr_write;
    }
    return nr_write / width;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This returns the file number associated with a stream.
 */
int fileno(FILE *file)
{
    CHECK(file);
    return file->fd;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This does absolutely nothing useful.
 */
int fflush(FILE *file)
{
    CHECK(file);
    file->eof = 0;
    return 0;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This clears the EOF indicator associated with a stream.
 */
void clearerr(FILE *file)
{
    VCHECK(file);
    file->eof = 0;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This returns the end-of-file indicator associated with a stream.
 */
int feof(FILE *file)
{
    CHECK(file);
    return file->eof;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  This returns the last error associated with a stream.
 */
int ferror(FILE *file)
{
    CHECK(file);
    return file->err;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This creates a child that runs a shell. If this is opened in
 *   "r" mode, we can read from the child's STDOUT_FILENO using a pipe.
 *   If this is opened in "w" mode, we can write to the child's
 *   STDIN_FILENO using a pipe.
 *   The original command is the text that would follow shell
 *   invocation as `/bin/sh -c COMMAND`. 0 reads 1 writes for pipes.
 */
FILE *popen(const char *cmd, const char *type)
{
    size_t i;
    int fd2close;
    struct sigaction sa;
    char *args[NR_ARGS];
    FILE *file;
    char buf[strlen(cmd)+1];

    if((cmd == NULL) || (type == NULL))
    {
        __set_errno(EINVAL);
        return NULL;
    }
    if(!((*type == (char)'r') || (*type == (char)'w')))
    {
        __set_errno(EINVAL);
        return NULL;
    }
    if((file = (FILE *) malloc(sizeof(FILE))) == NULL)
    {
        return file;
    }
    strcpy(buf, cmd);
    bzero(file, sizeof(FILE));
    if(pipe(file->pfd))
    {
        free(file);
        return NULL;
    }
    fd2close = 0xff;
    if(*type == (char)'r')
    {
        file->fd = file->pfd[0];
        fd2close = file->pfd[1];
    }
    else
    {
        file->fd = file->pfd[1];
        fd2close = file->pfd[0];
    }
    i = 0;
    args[i++] = "/bin/sh";
    args[i++] = "-c";
    args[i++] = buf;
    args[i] = NULL;
    sigaction(SIGCHLD, NULL, &sa);     /* Save old */
    signal(SIGCHLD, SIG_IGN);
    switch((file->pid = fork()))
    {
    case 0:
        if(*type == (char)'r')
        {
            dup2(file->pfd[1], STDOUT_FILENO);
            (void)close(file->pfd[0]);
        }
        else
        {
            dup2(file->pfd[0], STDIN_FILENO);
            (void)close(file->pfd[1]);
        }
        signal(SIGINT, SIG_IGN);
        signal(SIGQUIT, SIG_IGN);
        execve(args[0], args, __environ);
        exit(EXIT_FAILURE);
        break;
    case -1:
        (void)close(file->pfd[0]);
        (void)close(file->pfd[1]);
        free(file);
        return NULL;
    default:
        break;
    }
    file->magic = POPEN;
    sigaction(SIGCHLD, &sa, NULL);     /* Restore old */
    (void)close(fd2close);
    return file;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This closes the monster created above. Hopefully all resources
 *   are freed.
 */
int pclose(FILE *file)
{
    int status;
    CHECK(file);
    (void)kill(file->pid, SIGINT);
    (void)kill(file->pid, SIGTERM);
    if(wait4(file->pid, &status, WNOHANG, NULL) == FAIL)
        status = 0;
    (void)close(file->fd);
    free(file);
    return status;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
