//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
//
//  Copyright(c)  2005  Richard B. Johnson
//
//  This program may be distributed under the GNU Public License
//  version 2, as published by the Free Software Foundation, Inc.,
//  59 Temple Place, Suite 330 Boston, MA, 02111.
//
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
// File: stdio.c        Created 12-JAN-2005     Richard B. Johnson
//
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <malloc.h>
#include <errno.h>
#include <stdarg.h>
#include <ctype.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/wait.h>

extern void __set_errno(int);
extern int  __get_errno(void);
extern int __lldiv(long long *, int);
extern const char __small_num[];
extern const char __large_num[];
extern size_t strnlen(const char *string, size_t len);
#define FAIL -1
#define INT_MAX  0x7fffffff
#define MAX_STR  0x100           /* Maximum error string length   */
#define NR_ARGS  0x04            /* Number of args for 'system()' */

/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  This performs abnormal program termination and writes a core-
 *  dump.
 */
void __attribute__((__noreturn__))abort()
{
    kill(getpid(), SIGABRT);
    for(;;) ;
}
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
//
//   This clears the environment strings. It doesn't free them.
//
int clearenv()
{
   __environ[0] = NULL;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This searches for an environment string and if it exists, changes it
 *   if the overwrite parameter is true. It it does not exist, it creates it.
 *   In the event that it doesn't exist, the old string *env[] is never
 *   freed. Instead, and new one is created as shown in the `man` pages.
 */
int setenv(const char *name, const char *value, int overwrite)
{
    char *cp, **cpp;
    size_t i, name_len, value_len, len;
    name_len = strlen(name);
    value_len = strlen(value);
    if((cp = getenv(name)) != NULL)
    {
        if(overwrite)
        {
            if(strlen(value) <= strlen(cp))
                strcpy(cp, value);
            else
            {
               i = 0;
                cp -= (strlen(name) + 1);   /* Beginning of the string */
                while(__environ[i] != NULL)
                {
                    if(cp == __environ[i])
                    {
                        len = name_len + value_len + 0x04;
                        if((__environ[i] = (char *) malloc(len)) == NULL)
                        {
                            __set_errno(ENOMEM);
                            return FAIL;
                        }
                        strcpy(__environ[i], name);
                        strcat(__environ[i], "=");
                        strcat(__environ[i], value);
                        break;
                    }
                    i++;
                }
            }
        }
    }
    else
    {
        i = 0;
        len = name_len + value_len + 0x04;
        while(__environ[i++] != NULL)
            ;
        cpp = (char **) malloc(i * sizeof(char *));
        i = 0;
        while((cpp[i] =__environ[i]) != NULL)
            i++;
        if((cpp[i] = (char *) malloc(len)) == NULL)
        {
            __set_errno(ENOMEM);
            return FAIL;
        }
        strcpy(cpp[i], name);
        strcat(cpp[i], "=");
        strcat(cpp[i], value);
        cpp[i+1] = NULL;
        __environ = cpp;
    }
    return 0;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This is the same as above, but it __will__ make a new environment
 *   string if one doesn't exist.
 */
int putenv(char *string)
{
    int ret;
    char *nam;
    char *cp;
    if((cp = strchr(string, '=')) == NULL)
    {
        __set_errno(EINVAL);
        return FAIL;
    }
    if((nam = strdup(string)) == NULL)
    {
        __set_errno(ENOMEM);
        return FAIL;
    }
    cp = strchr(nam, '=');
    *cp++ = (char) 0x00;
    ret = setenv(nam, cp, 1);
    free(nam);
    return ret;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This spawns a shell to execute a command. Refer to the `man`
 *   pages for further info.
 */

int system(const char *command)
{
    size_t i;
    int status;
    struct sigaction sa;
    char *args[NR_ARGS];
    char cmd[strlen(command)+1];
    strcpy(cmd, command);
    i = 0;
    args[i++] = "/bin/sh";
    args[i++] = "-c";
    args[i++] = cmd;
    args[i] = NULL;
    sigaction(SIGCHLD, NULL, &sa);     /* Save old */
    signal(SIGCHLD, SIG_IGN);
    switch(fork())
    {
    case 0:
        signal(SIGINT, SIG_IGN);
        signal(SIGQUIT, SIG_IGN);
        execve(args[0], args, __environ);
        exit(EXIT_FAILURE);
        break;
    case -1:
        return FAIL;
    default:
        wait(&status);
    }
    sigaction(SIGCHLD, &sa, NULL);     /* Restore old */
    return status >> 8;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  This writes the input string, followed by a ':', followed by
 *  the text relating to the contents of the errno global variable
 *  to standard error.
 */

void perror(const char *cp)
{
    size_t len = strlen(cp) + MAX_STR;
    do {
    char buf[len];
    write(STDERR_FILENO, buf, sprintf(buf,"%s: %s\n", cp,
                  strerror(__get_errno())));
    } while(0);
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 * simple_strtoul - convert a string to an unsigned long
 * cp = the start of the string
 * endp = pointer to the end of the parsed string will be placed here
 * base = number base
 */
unsigned long simple_strtoul(const char *cp,char **endp,unsigned int base)
{
    unsigned long result = 0, value;
    if(!base)
    {
        base = 10;
        if(*cp == '0')
        {
            base = 8;
            cp++;
            if((*cp == 'x') && isxdigit(cp[1]))
            {
                cp++;
                base = 16;
            }
        }
    }
    while (isxdigit(*cp) &&
               (value = isdigit(*cp) ? *cp-'0' : toupper(*cp)-'A'+10) < base)
    {
        result = (result * base) + value;
        cp++;
    }
    if(endp)
        *endp = (char *)cp;
    return result;
}

/*
 * simple_strtol  Convert a string to a signed long
 */
long simple_strtol(const char *cp,char **endp,unsigned int base)
{
    if(*cp=='-')
        return -simple_strtoul(cp+1,endp,base);
    return simple_strtoul(cp,endp,base);
}

/*
 * simple_strtoull Convert a string to an unsigned long long
 */
unsigned long long simple_strtoull(const char *cp,char **endp,unsigned int base)
{
        unsigned long long result = 0,value;

        if (!base) {
                base = 10;
                if (*cp == '0') {
                        base = 8;
                        cp++;
                        if ((*cp == 'x') && isxdigit(cp[1])) {
                                cp++;
                                base = 16;
                        }
                }
        }
        while (isxdigit(*cp) && (value = isdigit(*cp) ? *cp-'0' : (islower(*cp)
            ? toupper(*cp) : *cp)-'A'+10) < base) {
                result = result*base + value;
                cp++;
        }
        if (endp)
                *endp = (char *)cp;
        return result;
}

/*
 * simple_strtoll  Convert a string to a signed long long
 */
long long simple_strtoll(const char *cp,char **endp,unsigned int base)
{
        if(*cp=='-')
                return -simple_strtoull(cp+1,endp,base);
        return simple_strtoull(cp,endp,base);
}

static int skip_atoi(const char **s)
{
        int i=0;

        while (isdigit(**s))
                i = i*10 + *((*s)++) - '0';
        return i;
}
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
//  Format a string and place it in a buffer if a va_list.
//
int vsprintf(char *buf, const char *fmt, va_list args)
{
    return vsnprintf(buf, 0xFFFFFFFFUL, fmt, args);
}
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
//  Format a string and place it in a buffer
//
int sprintf(char * buf, const char *fmt, ...)
{
        va_list args;
        int i;

        va_start(args, fmt);
        i=vsprintf(buf,fmt,args);
        va_end(args);
        return i;
}
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-

/*
 * vsscanf  Unformat a buffer into a list of arguments
 */
int vsscanf(const char * buf, const char * fmt, va_list args)
{
        const char *str = buf;
        char *next;
        int num = 0;
        int qualifier;
        int base;
        int field_width = -1;
        int is_sign = 0;

        while(*fmt && *str) {
                /* skip any white space in format */
                /* white space in format matchs any amount of
                 * white space, including none, in the input.
                 */
                if (isspace(*fmt)) {
                        while (isspace(*fmt))
                                ++fmt;
                        while (isspace(*str))
                                ++str;
                }

                /* anything that is not a conversion must match exactly */
                if (*fmt != '%' && *fmt) {
                        if (*fmt++ != *str++)
                                break;
                        continue;
                }

                if (!*fmt)
                        break;
                ++fmt;

                /* skip this conversion.
                 * advance both strings to next white space
                 */
                if (*fmt == '*') {
                        while (!isspace(*fmt) && *fmt)
                                fmt++;
                        while (!isspace(*str) && *str)
                                str++;
                        continue;
                }

                /* get field width */
                if (isdigit(*fmt))
                        field_width = skip_atoi(&fmt);

                /* get conversion qualifier */
                qualifier = -1;
                if (*fmt == 'h' || *fmt == 'l' || *fmt == 'L' || *fmt == 'Z') {
                        qualifier = *fmt;
                        fmt++;
                }
                base = 10;
                is_sign = 0;

                if (!*fmt || !*str)
                        break;

                switch(*fmt++) {
                case 'c':
                {
                        char *s = (char *) va_arg(args,char*);
                        if (field_width == -1)
                                field_width = 1;
                        do {
                                *s++ = *str++;
                        } while(field_width-- > 0 && *str);
                        num++;
                }
                continue;
                case 's':
                {
                        char *s = (char *) va_arg(args, char *);
                        if(field_width == -1)
                                field_width = INT_MAX;
                        /* first, skip leading white space in buffer */
                        while (isspace(*str))
                                str++;

                        /* now copy until next white space */
                        while (*str && !isspace(*str) && field_width--) {
                                *s++ = *str++;
                        }
                        *s = '\0';
                        num++;
                }
                continue;
                case 'n':
                        /* return number of characters read so far */
                {
                        int *i = (int *)va_arg(args,int*);
                        *i = str - buf;
                }
                continue;
                case 'o':
                        base = 8;
                        break;
                case 'x':
                case 'X':
                        base = 16;
                        break;
                case 'd':
                case 'i':
                        is_sign = 1;
                case 'u':
                        break;
                case '%':
                        /* looking for '%' in str */
                        if (*str++ != '%')
                                return num;
                        continue;
                default:
                        /* invalid format; stop here */
                        return num;
                }

                /* have some sort of integer conversion.
                 * first, skip white space in buffer.
                 */
                while (isspace(*str))
                        str++;

                if (!*str)
                    break;
                if( (!isdigit(*str)) && (!isxdigit(*str)) )
                        break;

                switch(qualifier) {
                case 'h':
                        if (is_sign) {
                                short *s = (short *) va_arg(args,short *);
                                *s = (short) simple_strtol(str,&next,base);
                        } else {
                                unsigned short *s = (unsigned short *) va_arg(args, unsigned short *);
                                *s = (unsigned short) simple_strtoul(str, &next, base);
                        }
                        break;
                case 'l':
                        if (is_sign) {
                                long *l = (long *) va_arg(args,long *);
                                *l = simple_strtol(str,&next,base);
                        } else {
                                unsigned long *l = (unsigned long*) va_arg(args,unsigned long*);
                                *l = simple_strtoul(str,&next,base);
                        }
                        break;
                case 'L':
                        if (is_sign) {
                                long long *l = (long long*) va_arg(args,long long *);
                                *l = simple_strtoll(str,&next,base);
                        } else {
                                unsigned long long *l = (unsigned long long*) va_arg(args,unsigned long long*);
                                *l = simple_strtoull(str,&next,base);
                        }
                        break;
                case 'Z':
                {
                        size_t *s = (size_t*) va_arg(args,size_t*);
                        *s = (size_t) simple_strtoul(str,&next,base);
                }
                break;
                default:
                        if (is_sign) {
                                int *i = (int *) va_arg(args, int*);
                                *i = (int) simple_strtol(str,&next,base);
                        } else {
                                unsigned int *i = (unsigned int*) va_arg(args, unsigned int*);
                                *i = (unsigned int) simple_strtoul(str,&next,base);
                        }
                        break;
                }
                num++;

                if (!next)
                        break;
                str = next;
        }
        return num;
}

//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
//  Unformat a buffer into a list of arguments
//
int sscanf(const char * buf, const char * fmt, ...)
{
        va_list args;
        int i;

        va_start(args,fmt);
        i = vsscanf(buf,fmt,args);
        va_end(args);
        return i;
}
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-


